"""A file containing the config and code for interacting with Bucket Service."""

import os
import asyncio
import tarfile
from foglamp.common.bucket_client.model_query_builder import ModelQueryBuilder

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"


ae_bucket_config_properties = {
    "constant": {
        "type": "model"
    },
    "key": {
        "name": {
            "description": "Archive containing models that is to be picked from bucket service.",
            "type": "string",
            "default": "Model archive",
            "order": "1",
            "displayName": "Model archive name"
        },
    },
    "properties": {
        "version": {
            "description": "Archive version as stored in bucket",
            "type": "string",
            "default": "1.0",
            "order": "2",
            "displayName": "Model version"
        },
        "scaler": {
            "description": "The file name of scaler that will be loaded from the archive.",
            "type": "string",
            "default": "ae_min_max_scaler.pkl",
            "order": "3",
            "displayName": "Scaler",
            "readonly": "true"
        },
        "ml_model": {
            "description": "The file name of ML model that will be loaded from the archive.",
            "type": "string",
            "default": "adm_ld_ae_float32.tflite",
            "order": "4",
            "displayName": "ML model",
            "readonly": "true"
        },
        "model_metadata": {
            "description": "The file name of model metadata that will be loaded from the archive.",
            "type": "string",
            "default": "ae_model_data.npy",
            "order": "5",
            "displayName": "Model metadata",
            "readonly": "true"
        }
    }
}

ae_bucket_config_default = {
    "type": ae_bucket_config_properties["constant"]["type"],
    "name": ae_bucket_config_properties["key"]["name"]["default"],
    "version": ae_bucket_config_properties["properties"]["version"]["default"],
    "scaler": ae_bucket_config_properties["properties"]["scaler"]["default"],
    "ml_model": ae_bucket_config_properties["properties"]["ml_model"]["default"],
    "model_metadata": ae_bucket_config_properties["properties"]["model_metadata"]["default"],
}


def get_model_files(
    logger, bucket_client, bucket_type, model_name, model_version
):
    """Get model archive file from the bucket service and extract it at a path.

    FUTURE: how about loading the model directly without writing to disk
    Setting up paths
        - model_dir directory: where the artifacts will be extracted
        - archive saving path: where archive coming from bucket service will be
                               saved

    Args:
        bucket_client: Bucket client handle to use bucket services
        bucket_type: type defined in the attributes in bucket.
            it should match with
            ae_bucket_config_properties.constant.type --> "adm_ld_filter_models"
        model_name: Model name e.g. "TFLite AE Model Archive"
        model_version: Model's version e.g. "1.0"
    Returns:
        Model directory: where the models were extracted.
    Raises:
    """
    # Paths setup
    model_dir = os.path.join("/tmp", "foglamp_adm_ld_filter")
    archive_saving_path = os.path.join(model_dir, "adm_ld_filter_archive.tar.gz")

    # Prepare the directory
    os.system(f"rm -rf {model_dir}")
    os.makedirs(model_dir)

    # creating query for searching the bucket using the attributes
    model_attrs = {
        "type": bucket_type,
        "name": model_name,
        "version": model_version,
    }
    mqb = ModelQueryBuilder()
    for param, user_defined_value in model_attrs.items():
        mqb = mqb.MATCH([param, "=", user_defined_value])
    query = mqb.payload()
    logger.info(f"bucket_client.query_models: query payload={query}")

    # Create a temporary event loop for async calls in this function since
    # sometimes this function gets called before event loop creation
    # (plugin_init) and sometimes after (plugin_reconfigure)
    loop = asyncio.new_event_loop()

    # Locating the matching bucket
    try:
        # getting the matching bucket options based on the query created above
        task = asyncio.ensure_future(
            bucket_client.query_models(query), loop=loop
        )
        loop.run_until_complete(task)
        result = task.result()
        logger.info(f"bucket_client.query_models returned {result}")

        # checking the matches
        if len(result["matches"]) < 1:
            logger.error("Didn\'t find any matching data within bucket service.")
            return None
        if len(result["matches"]) == 1:
            logger.info("Found the exact match within bucket service. Loading it.")
        if len(result["matches"]) > 1:
            logger.warn("Found more than one match within bucket service. Picking first one.")

        # loading the match
        result = result["matches"][0]
        logger.info(f"Chosen model : {result}")

    except Exception as ex:
        _msg = "ADM LD filter is unable to get a match within bucket service."
        _msg += f"\nDetails: {str(ex)}"
        logger.error(_msg)
        return None

    # Get the data from the bucket & save it to the mentioned path
    task = asyncio.ensure_future(
        bucket_client.get_bucket_content(result["bucket"], archive_saving_path),
        loop=loop
    )
    loop.run_until_complete(task)
    result = task.result()
    loop.stop()
    loop.close()
    logger.info(f"Saved the archive from the bucket to {archive_saving_path}")

    # Extract the archive to the model directory
    logger.info(f"Un-compressing archive {archive_saving_path} in {model_dir}")
    try:
        tar = tarfile.open(archive_saving_path, "r|*")
        tar.extractall(model_dir)
        tar.close()
    except Exception as ex:
        logger.error(
            "Unable to extract model files from (assumed) compressed archive:" +
            f"{result['file']}, exception={str(ex)}"
        )
        return None

    return model_dir
