# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Server functions for plugin module to detect need for large discharge (LD) """


__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"

import os
import gc
import copy
import datetime
import logging

import pandas as pd

from foglamp.common import logger, utils
from foglamp.plugins.filter.ADM_LD_prediction import ml_eval

# canonical data field names
KEY_DATA1 = 'Channel0'
KEY_DATA2 = 'Channel1'
KEY_DATA3 = 'Channel2'
KEY_DATA4 = 'Channel3'
KEY_TS = 'user_ts'


# very special: if we are passed in a default logger destination use that
if hasattr(logger, 'FOGLAMP_DESTINATION_ENV') and (logger.FOGLAMP_DESTINATION_ENV in os.environ):
    logger.set_default_destination(int(os.environ[logger.FOGLAMP_DESTINATION_ENV]))
else:
    utils.eprint("NO DEST ENV")

_LOGGER = logger.setup(__name__, level=logging.INFO)
LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR
}

@utils.add_functions_as_methods(ml_eval.export_service_functions)
class ModelInferenceHandler:

    def __init__(self):

        MAX_FRAMES = 10   # number seconds of data for model evaluation

        self.logger = _LOGGER
        
        self.handle = dict()
        
        # data are input into an in-memory frame
        self.df = pd.DataFrame()
        self.num_rows_processed = 0
        self.num_input_rows = 0
        self.num_frames = 0
        self.frame = None
        self.frames = []
        self.max_frames = MAX_FRAMES
        
        self.pca = None
        self.min_max_scaler = None
        # self.yeo_johnson_transform = None
        self.interpreter = None
        self.model_data = None

        self.tpu = False
        self.forwardData = False
        self.valid = False

        self.do_check_anomaly = True
        self.sampling_rate = 8000
        self.frame_size = self.sampling_rate * 5
        self.inference_gap = None
        self.cols = [KEY_TS,  KEY_DATA1, KEY_DATA2, KEY_DATA3]

        self.lastProcessedDataTS = pd.Timestamp('2000-01-01 00:00:00.000000+00:00')

    def initialize(self, handle):
        self.logger.info('PLUGIN INIT')
        self.logger.setLevel(LOGLEVEL[handle['logLevel']['value']])

        handle = copy.deepcopy(handle)

        # initialize data and common settings
        _sampling_rate = handle['samplingRate']['value']
        if _sampling_rate.endswith('khz'):
            self.sampling_rate = int(_sampling_rate[:-3])*1000
        else:
            self.sampling_rate = int(_sampling_rate)
        self.frame_size = self.sampling_rate*5

        self.forwardData = (handle['forwardData']['value'] == 'true')
        self.inference_gap = pd.to_timedelta(handle['evaluationInterval']['value'])

        # initialize ML-specific values
        handle = self.ml_initialize(handle)

        self.handle = handle
        self.valid = True
        return handle

    def shutdown(self):
        self.logger.warning("ModelServer: SHUTTING DOWN")
        self.valid = False

    def evaluate(self, data):
        """ Evaluates whether a large discharge is required by running vibration data thru' ML model
        Args:
            data: 2 array of readings:
            {'channel1': [amp1, amp2, ...]  
             'ts': [ts1, ts2, ...]}
        Returns:
            Returns a bool indicating whether LD is required
        """
        self.logger.debug("ModelServer: plugin_ingest")
        handle = self.handle

        if not self.valid:
            self.logger.error("Cannot process data since NN interpreter is not available")
            return -1, [], []

        try:
            self.logger.debug("srvev: "+str(data))

            if not self.accumulate_data(data):
                return -1, [], []

            anomalous, RE_list, smoothed_RE_list = self.check_anomaly(handle)
            return anomalous, RE_list, smoothed_RE_list
        
        except Exception as ex:
            self.logger.exception('Error in ModelServer plugin_ingest {}'.format(str(ex)))
            self.valid = False
            raise

    def accumulate_data(self, data):
        """ accumulate_data - add data to the dataframe we assemble for analysis """

        # clean data: convert to canonical names
        if len(data) == 0:
            return []

        t1 = datetime.datetime.now()

        _input = pd.DataFrame(data[0])
        currTS = pd.to_datetime(_input['user_ts'][0])  # TS in first reading
        if currTS < self.lastProcessedDataTS + self.inference_gap:
            self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so NOT processing this data frame"
                              .format(self.lastProcessedDataTS, self.inference_gap, currTS))

            del data, _input
            # gc.collect()
            return False

        t1 = datetime.datetime.now()
        # The following code assumes that the data coming to the filter
        # is a list of dicts, where each dict is not a nested dict.
        # If nested dicts are there then use the following call.
        #  _input = _input.join(pd.json_normalize(_input['readings']))
        required_dict_list = [s['readings'] for s in data]
        required_ts_list = [s['user_ts'] for s in data]
        _input_d = pd.DataFrame(required_ts_list, columns=['user_ts_orig'])
        _input_r = pd.DataFrame.from_records(required_dict_list)
        _input = pd.concat([_input_r, _input_d], axis=1)
        extra_cols = list(_input.columns)
        extra_cols.remove('user_ts_orig')

        extra_cols = set(extra_cols) - set(self.cols)
        self.logger.debug("self.cols={}, extra_cols = {}".format(self.cols, extra_cols))
        if len(extra_cols):
            _input.drop(labels=list(extra_cols), axis=1, inplace=True)

        _input.index = pd.to_datetime(_input['user_ts_orig'])
        labels_to_remove = ['user_ts_orig']
        if 'user_ts' in _input.columns:
            labels_to_remove.append('user_ts')
        _input.drop(labels=labels_to_remove, axis=1, inplace=True)
        
        t2 = datetime.datetime.now()
        delta = t2 - t1
        delta = int(delta.total_seconds() * 1000 * 1000)  # in usec
        self.logger.info(
            "df processing took {:.0f} msec for {} readings, time/reading={:.0f} us"
                .format(delta/1000, _input.shape[0], delta/_input.shape[0]))

        self.logger.debug("lastProcessedDataTS={}, inference_gap={}, currTS={}, so processing this data frame"
                          .format(self.lastProcessedDataTS, self.inference_gap, currTS))

        _len = _input.shape[0]
        self.df = pd.concat([self.df, pd.DataFrame(_input)], ignore_index=True)
        self.num_input_rows += _len

        if self.df.shape[0] < self.frame_size:
            self.logger.debug("###### Got {} rows to process, not processing yet".format(self.df.shape[0]))
            return False   # don't run model inference if data length is not sufficient

        self.lastProcessedDataTS = currTS

        self.num_rows_processed += _len
        self.num_frames += 1

        self.frame = self.df[-self.frame_size:]

        self.df = pd.DataFrame()

        return True
