# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Plugin module to extract WPT features from vibration data """

import numpy as np
import pandas as pd

import scipy
from scipy import fftpack, signal
from scipy.stats import kurtosis
from scipy.stats import skew
from collections import Counter

import pywt

from foglamp.common import logger

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"
__email__ = "info@dianomic.com"

_LOGGER = logger.setup(__name__)
CHANNEL_LIST = ["Channel0", "Channel1", "Channel2"]


def _kurtosis(x):
    return kurtosis(x)


def CPT5(x):
    den = len(x) * np.exp(np.std(x))
    return sum(np.exp(x)) / den


def SSC(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    xn_i1 = x[0:len(x) - 2]  # xn-1
    ans = np.heaviside((xn - xn_i1) * (xn - xn_i2), 0)
    return sum(ans[1:])


def wave_length(x):
    x = np.array(x)
    x = np.append(x[-1], x)
    x = np.append(x, x[1])
    xn = x[1:len(x) - 1]
    xn_i2 = x[2:len(x)]  # xn+1
    return sum(abs(xn_i2 - xn))


def norm_entropy(x):
    tresh = 2
    return sum(np.power(abs(x), tresh))


def SRAV(x):
    SRA = sum(np.sqrt(abs(x)))
    return np.power(SRA / len(x), 2)


def mean_abs(x):
    return sum(abs(x)) / len(x)


def calculate_entropy(list_values):
    counter_values = Counter(list_values).most_common()
    probabilities = [elem[1] / len(list_values) for elem in counter_values]
    entropy = scipy.stats.entropy(probabilities)
    return entropy


def calculate_statistics(list_values):
    n5 = np.nanpercentile(list_values, 5)
    n25 = np.nanpercentile(list_values, 25)
    n75 = np.nanpercentile(list_values, 75)
    n95 = np.nanpercentile(list_values, 95)
    median = np.nanpercentile(list_values, 50)
    mean = np.nanmean(list_values)
    std = np.nanstd(list_values)
    var = np.nanvar(list_values)
    rms = np.nanmean(np.sqrt(list_values ** 2))
    # New features
    kur = kurtosis(list_values)
    MeanAbs = mean_abs(list_values)
    norm_ent = norm_entropy(list_values)
    skewness = skew(list_values)
    CPT_5 = CPT5(list_values)
    SSC_1 = SSC(list_values)
    WL = wave_length(list_values)
    SRAV_1 = SRAV(list_values)
    return [n5, n25, n75, n95, median, mean, std, var, rms, kur, MeanAbs, norm_ent, skewness, CPT_5, SSC_1, WL, SRAV_1]


def calculate_crossings(list_values):
    zero_crossing_indices = np.nonzero(np.diff(np.array(list_values) > 0))[0]
    no_zero_crossings = len(zero_crossing_indices)
    mean_crossing_indices = np.nonzero(np.diff(np.array(list_values) > np.nanmean(list_values)))[0]
    no_mean_crossings = len(mean_crossing_indices)
    return [no_zero_crossings, no_mean_crossings]


def get_features(list_values):
    entropy = calculate_entropy(list_values)
    crossings = calculate_crossings(list_values)
    statistics = calculate_statistics(list_values)
    return [entropy] + crossings + statistics


wavelet = pywt.Wavelet('db20')  # mother wavelet to use
level = 4  # level of Wavelet Packet decomposition


def get_wpt_features(time_series):
    """ Do the Wavelet packet decomposition and extract features
    Args:
        time_series: vibration data
    Returns:
        features: WPT features list for 1 interval of rows
    """
    features = []
    wptree = pywt.WaveletPacket(time_series, wavelet, mode='symmetric', maxlevel=level)
    levels = wptree.get_level(level, order="freq")
    for node in levels:
        data_wp = node.data
        features.extend(get_features(data_wp))
    # print(len(features))
    return features


# Velocity features
def get_vel_stats(time_series, fs):
    _sig = signal.detrend(time_series)
    #     print("_sig.shape={}, _sig[:5]={}".format(_sig.shape, _sig[:5]))

    amps = fftpack.fft(_sig)
    freqs = fftpack.fftfreq(_sig.shape[0], d=1 / fs)
    fft_bin_width = freqs[1]
    skip_bins = int(round((fs / 1000) / fft_bin_width))

    amps = amps / (2 * np.pi * (1j) * (freqs + [fft_bin_width / 2]))
    amps[:skip_bins] = [0]
    amps[-skip_bins:] = [0]

    vel = fftpack.ifft(amps)

    vel = np.real(vel)
    vel = vel * [1000]  # convert from m/s to mm/s

    features = get_features(vel)

    return features

def parse_vibration_data(df, sampling_rate, interval_rows, num_channels):
    """ Get WPT features for each set of 'interval_rows' rows
    Args:
        df: input vibration data
        interval_rows: Number of data rows over which features are calculated
        num_channels: Number of channels of vibration data
    Returns:
        feature_list: list with WPT features
    """
    # _LOGGER.info("type(df)={}, len(df)={}, df.shape={}".format(type(df), len(df), df.shape))
    # _LOGGER.info("type(df[0])={}, df[0].shape={}".format(type(df[0]), df[0].shape))

    df = pd.DataFrame(df)
    # _LOGGER.info("parse_vibration_data(): df.shape={}".format(df.shape))
    df.columns = ["Channel0", "Channel1", "Channel2"]
    feature_list = []
    for r in range(0, df.shape[0], interval_rows):
        features = []
        if r % 1000 == 0 and r != 0:
            print("{} rows done".format(r))
        start = r
        end = min(r + interval_rows, df.shape[0])
        # _LOGGER.info("chunk range={}..{}".format(start, end))
        # features.extend(get_wpt_features(df['channel1'][start:end]))
        # if num_channels == 2:
        #     features.extend(get_wpt_features(df['channel2'][start:end]))
        for ch in CHANNEL_LIST:
            features.extend(get_wpt_features(df[ch][start:end]))
        for ch in CHANNEL_LIST:
            features.extend(get_vel_stats(df[ch][start:end], sampling_rate))
        feature_list.append(features)

    return feature_list
